/*
** Module   : WidgetBrowser.java
** Abstract : Implements Widget Browser UI.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20141031 Created initial version
*/
package com.goldencode.p2j.ui.client.widgetbrowser;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.border.EmptyBorder;
import javax.swing.tree.TreePath;

import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.ui.client.widget.Container;

/**
 * Implements Widget Browser UI.
 */
public class WidgetBrowser
{
   /**
    * The main window.
    */
   private JFrame windowFrame;
   
   /**
    * The tree component for displaying widget hierarchy.
    */
   private ContainerTree tree;

   /**
    * The constructor.
    * 
    * @param   container
    *          The root container, usually a window reference.
    *          
    * @param   actionListener
    *          An optional events listener, see {@link Event} for available events.
    */
   public WidgetBrowser(final Container<?> container, final ActionListener actionListener) 
   {
      // build main window title
      String windowTitle = "<" + container.config().id + ">" + 
                           (container.name() != null ? container.name() : "");
      windowTitle = String.format("Widget Browser %s", windowTitle);
      
      // main window
      windowFrame = new JFrame(windowTitle);
      windowFrame.setSize(900, 600);
      windowFrame.setLocation(100, 100);
      windowFrame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
      windowFrame.addWindowListener(new WindowListener()
      {
         @Override
         public void windowOpened(WindowEvent e)
         {
         }

         @Override
         public void windowClosing(WindowEvent e)
         {
         }

         @Override
         public void windowClosed(WindowEvent e)
         {
            if (actionListener != null)
            {
               actionListener.actionPerformed(new Event(this, Event.BROWSER_CLOSED, container));
            }
         }

         @Override
         public void windowIconified(WindowEvent e)
         {
         }

         @Override
         public void windowDeiconified(WindowEvent e)
         {
         }

         @Override
         public void windowActivated(WindowEvent e)
         {
         }

         @Override
         public void windowDeactivated(WindowEvent e)
         {
         }
      });

      // add content panel
      JPanel contentPanel = new JPanel();
      contentPanel.setBorder(new EmptyBorder(5, 5, 5, 5));
      contentPanel.setLayout(new BorderLayout(0, 0));
      windowFrame.setContentPane(contentPanel);

      // property table
      PropertyTable table = new PropertyTable();
      final ContainerTree tree = new ContainerTree(container, table);
      tree.setActionListener(actionListener);

      // table event handler
      table.setActionListener(new ActionListener() 
      {
         @Override
         public void actionPerformed(ActionEvent paramActionEvent) 
         {
            if (paramActionEvent.getID() != Event.PROPERTY_VALUE_NAVIGATED)
            {
               return;
            }
            
            Object propValue = ((Event)paramActionEvent).propertyValue;
            if (propValue instanceof Widget<?>)
            {
               TreePath result = tree.findNode((Widget<?>) propValue);
               if (result != null) {
                  tree.setSelectionPath(result);
               }
            }
         }
      });
      
      this.tree = tree;

      // add top panel for buttons
      JPanel panelTop = new JPanel();
      contentPanel.add(panelTop, BorderLayout.NORTH);
      panelTop.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));

      // refresh button
      JButton refreshButton = new JButton("Refresh");
      panelTop.add(refreshButton);
      refreshButton.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent e) 
         {
            tree.refresh();
         }
      });

      // collapse all button
      JButton collapseAllButton = new JButton("Collapse All");
      panelTop.add(collapseAllButton);
      collapseAllButton.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent e) 
         {
            tree.collapseAll();
         }
      });

      // expand all button
      JButton expandAllButton = new JButton("Expand All");
      panelTop.add(expandAllButton);
      expandAllButton.addActionListener(new ActionListener() 
      {
         public void actionPerformed(ActionEvent e) 
         {
            tree.expandAll();
         }
      });

      // full class names check-box
      final JCheckBox showFullClassName = new JCheckBox("Full Class Names");
      showFullClassName.setSelected(tree.isShowFullClassName());
      showFullClassName.addItemListener(new ItemListener() 
      {
         public void itemStateChanged(ItemEvent e) 
         {
            tree.setShowFullClassName(showFullClassName.isSelected());
            tree.refresh();
         }
      });
      panelTop.add(showFullClassName);
      
      // add splitter
      JSplitPane splitPane = new JSplitPane();
      splitPane.setDividerLocation(300);
      contentPanel.add(splitPane, BorderLayout.CENTER);
      splitPane.setLeftComponent(new JScrollPane(tree));
      splitPane.setRightComponent(new JScrollPane(table));
   }
   
   /**
    * Shows the main window.
    */
   public void show()
   {
      windowFrame.setVisible(true);
   }
   
   /**
    * Returns <code>true</code> is the supplied widget is selected to be highlighted.
    * 
    * @param   widget
    *          A {@link Widget} reference.
    * 
    * @return  <code>true</code> if the supplied widget is selected to be highlighted,
    *          <code>false</code> otherwise.
    */
   public boolean isHighlighted(Widget<?> widget)
   {
      return widget != null && tree.isSelected(widget);
   }
}
