/*
** Module   : ProcedureManager.java
** Abstract : Implements access to procedure system handles and procedure
**            related attributes and methods. 
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130130 Added stubbed methods to implement the PUBLISH, SUBSCRIBE and UNSUBSCRIBE 
**                  language statements.
** 003 CA  20130131 Fixed problem determining THIS-PROCEDURE when the trigger invoked is defined 
**                  as ON ... PERSISTENT RUN.
** 004 CA  20130202 Update from character to TextOps.
** 005 CA  20130529 Some refactoring related to procedure deletion and added appserver support.
** 006 CA  20130918 Added runtime support for widget-pools: the pools need scope notifications,
**                  and this must be done by the ProcedureManager, to ensure proper access to the 
**                  THIS-PROCEDURE handle.
** 007 CA  20130925 Added support for named events.
** 008 CA  20131013 Implemented implicit deletion of static resources and delayed processing of
**                  finalizable/scopeable instances. Added support for the INSTANTIATING-PROCEDURE 
**                  attribute.
** 009 CA  20131208 Emit SEARCH-TARGET and SEARCH-SELF literals as integer.  Fixed runtime for 
**                  RUN SUPER and SUPER() calls, in search-target and search-self modes.
** 010 HC  20131222 Implemented ADM-DATA and UNIQUE-ID attribute.
**                  INSTANTIATING-PROCEDURE is now managed by ProcedureManager.
** 011 EVK 20131223 Made method isProcedure public.
** 012 CA  20140109 WorkArea.scopeFinished has no target-procedure to pop, when terminating an 
**                  appserver request.
** 013 VIG 20140109 Implemented CURRENT-WINDOW attribute.
** 014 CA  20140218 Implemented PROGRAM-NAME. Fixed some SOURCE- and TARGET-PROCEDURE cases (the
**                  RUN PERSISTENT or RUN ext-prog cases).  Also, the external program name used
**                  at the RUN statement must be used in error messages; the absolute name must be
**                  used each time SourceNameMapper APIs are invoked.  Now we allow java-style
**                  calls for both internal and external procedures.
** 015 CA  20140221 Fixed SUPER-PROCEDURES - emit correct handle ID.
** 016 CA  20140223 Fixed deletion of persistent procedures via DELETE OBJECT <system-proc-handle>.
** 017 CA  20140301 Procedure's scope is notified first for start and last for finish.
** 018 HC  20140424 Fixed resolveClosestMethod failing to resolve extent functions.
** 019 CA  20140502 Fixed super-procedure chaining in SUPER and RUN SUPER calls.
** 020 SVL 20140414 Modified INSTANTIATING-PROCEDURE assignment logic in order to support agent
**                  contexts.
** 021 CA  20140513 Added a weight for the context-local var, to ensure predetermined order during
**                  context reset.
** 022 HC  20140629 Implemented access to the internal callee information structure.
** 023 CA  20140729 Made getRelativeName public.
** 024 VIG 20140810 Added invalid param processing for setCurrentWindow method.
** 025 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 026 EVL 20141024 The dynamic flag need to be set early enough for dynamic widgets to avoid NPE
**                  in processResource() method.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.persist.StaticTempTable;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.UniqueIdGenerator.IdKind;

/**
 * Provides implementation for the following procedure handles:
 * <table>
 * <tr>
 *   <td>THIS-PROCEDURE</td>
 *   <td>{@link #thisProcedure()}</td>
 * </tr>
 * <tr>
 *   <td>SOURCE-PROCEDURE</td>
 *   <td>{@link #sourceProcedure()}</td>
 * </tr>
 * <tr>
 *   <td>TARGET-PROCEDURE</td>
 *   <td>{@link #targetProcedure()}</td>
 * </tr>
 * </table>
 * The following procedure-related attributes and methods are implemented:
 * <table>
 * <tr>
 *    <th>attribute or method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>SESSION:SUPER-PROCEDURES</td>
 *    <td>{@link #superProcedures()}</td>
 * </tr>
 * <tr>
 *    <td>SESSION:ADD-SUPER-PROCEDURE</td>
 *    <td>
 *       <ul>
 *          <li>{@link #addSuperProcedure(handle)}</li>
 *          <li>{@link #addSuperProcedure(handle, SearchMode)}</li>
 *          <li>{@link #addSuperProcedure(handle, int)}</li>
 *    </td>
 * </tr>
 * <tr>
 *    <td>SESSION:REMOVE-SUPER-PROCEDURE</td>
 *    <td>
 *       <ul>
 *          <li>{@link #removeSuperProcedure(handle)}</li>
 *       </ul>
 *    </td>
 * </tr>
 * <tr>
 *    <td>SESSION:FIRST-PROCEDURE</td>
 *    <td>{@link #firstProcedure()}</td>
 * </tr>
 * <tr>
 *    <td>SESSION:LAST-PROCEDURE</td>
 *    <td>{@link #lastProcedure()}</td>
 * </tr>
 * </table>
 * <p>
 * The {@link #registerFunctionHandle(String, Resolvable)} is emitted at the
 * beginning of an external procedure, for any function defined 
 * <code>IN handle</code>.
 */
public class ProcedureManager
{
   /** Constant defining search mode: SEARCH-SELF */
   public static final int SEARCH_SELF = 1;
   
   /** Constant for socket search mode: SEARCH_TARGET */
   public static final int SEARCH_TARGET = 2;

   /** Marks the external program's root procedure - this is a P2J reserver name. */
   public static final String EXTERNAL_PROGRAM = "__P2J__EXTERNAL__PROGRAM__";

   /** Marks the trigger block - this is what 4GL reports for triggers. */
   public static final String TRIGGER_PROCEDURE = "USER-INTERFACE-TRIGGER";
   
   /** 
    * Defines search modes for the ADD-SUPER-PROCEDURE statement. Needed to
    * be able to identify the correct target in RUN SUPER and SUPER() cases.
    */
   public static enum SearchMode
   {
      SEARCH_SELF
      {
         @Override
         public String toString()
         {
            return "SEARCH-SELF";
         }
      },
      
      SEARCH_TARGET
      {
         @Override
         public String toString()
         {
            return "SEARCH-TARGET-PROCEDURE";
         }
      }
   };

   /** Stores context-local state variables. */
   private static ContextContainer work = new ContextContainer();

   /**
    * Register the handle expression for the given function.
    * 
    * @param    pname
    *           The function's legacy 4GL name.
    * @param    handleExpr
    *           The {@link Resolvable} instance used to determine its IN 
    *           handle target.
    */
   public static void registerFunctionHandle(String pname, 
                                             Resolvable handleExpr)
   {
      // always work with thisProcedure
      handle h = thisProcedure();
      Object referent = h.get();
      ProcedureData pp = locateProcedure(referent);
      pp.inHandleFunctions.put(pname, handleExpr);
   }
   
   /**
    * Get the THIS-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle thisProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.thisProcedures.peek();
      return new handle(locateProcedure(referent).phandle);
   }
   
   /**
    * Get the TARGET-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle targetProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.targetProcedures.peek();
      return new handle(locateProcedure(referent).phandle);
   }

   /**
    * Get the SOURCE-PROCEDURE system handle.
    * 
    * @return   See above.
    */
   public static handle sourceProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.sourceProcedures.peek();

      return referent == null ? new handle()
                              : new handle(locateProcedure(referent).phandle);
   }
   
   /**
    * Get the SESSION:LAST-PROCEDURE handle (unknown if no persistent 
    * procedures are loaded).
    * 
    * @return   See above.
    */
   public static handle lastProcedure()
   {
      WorkArea wa = work.get();
      if (wa.persistentProcedures.isEmpty())
      {
         return new handle();
      }

      Object referent = wa.persistentProcedures.peekLast();
      return new handle(locateProcedure(referent).phandle);
   }
   
   /**
    * Get the SESSION:FIRST-PROCEDURE handle (unknown if no persistent 
    * procedures are loaded).
    * 
    * @return   See above.
    */
   public static handle firstProcedure()
   {
      WorkArea wa = work.get();
      if (wa.persistentProcedures.isEmpty())
      {
         return new handle();
      }

      Object referent = wa.persistentProcedures.peekFirst();     
      return new handle(locateProcedure(referent).phandle);
   }
   
   /**
    * Get the SESSION:SUPER-PROCEDURES attribute.
    * 
    * @return   See above.
    */
   public static character superProcedures()
   {
      WorkArea wa = work.get();
      return superProceduresImpl(wa.getSessionSuperProcedures().values());
   }

   /**
    * Add the given handle as a super-procedure for the SESSION handle.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h)
   {
      return addSuperProcedure(h, (SearchMode) null);
   }

   /**
    * Add the given handle as a super-procedure for the SESSION handle, using
    * the explicit {@link SearchMode mode}.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    searchMode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h, int searchMode)
   {
      return addSuperProcedure(h, getSearchMode(searchMode));
   }

   /**
    * Add the given handle as a super-procedure for the SESSION handle, using
    * the explicit {@link SearchMode mode}.
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    searchMode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h, integer searchMode)
   {
      if (searchMode.isUnknown())
      {
         return addSuperProcedure(h, (SearchMode) null);
      }
      else
      {
         return addSuperProcedure(h, searchMode.intValue());
      }
   }

   /**
    * Add the given handle as a super-procedure for the SESSION handle, using
    * the search-mode set as a string (one of the <code>SEARCH-TARGET</code>
    * or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical addSuperProcedure(handle h, SearchMode mode)
   {
      return addSuperProcedureImpl(h, mode, null);
   }

   /**
    * Remove the given handle from the SESSION's list of super-procedures.
    *  
    * @param    h
    *           The procedure handle to be removed.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   public static logical removeSuperProcedure(handle h)
   {
      if (!h._isValid() || !isProcedure(h))
      {
         return new logical(false);
      }

      removeProcedureImpl(h, work.get().getSessionSuperProcedures().keySet());
   
      return new logical(true);
   }

   /**
    * Process the given resource by either setting its INSTANTIATING-PROCEDURE attribute 
    * immediately (in case of non-static resources) or add the static resource to the 
    * {@link WorkArea#pendingResources} set, to be processed when the next scope is entered.
    * 
    * @param    res
    *           The resource to be processed.
    * @param    dynamic
    *           Flag indicating if this is a dynamic or static resource.
    */
   public static void processResource(WrappedResource res, boolean dynamic)
   {
      // be a no-op in case of headless mode
      if (ErrorManager.isHeadless())
      {
         return;
      }
      
      WorkArea wa = work.get();
      
      boolean pending = !(res instanceof StaticTempTable);
      
      if (res instanceof DynamicResource && !dynamic && pending)
      {
         // all static resources are created as instance fields; thus, the actual processing is 
         // delayed, as the THIS-PROCEDURE is not available at the time of the call (i.e. we are
         // in an instance field initialization code, at the business logic class)
         wa.pendingResources.add(res);
      }
      else if (res instanceof HandleResource)
      {
         wa.pendingResources.remove(res);
         
         // all resources which don't have dynamic counterparts are processed immediately
         Object referent = _thisProcedure();
         ProcedureData pdata = locateProcedure(referent);

         // make sure the resource is registered as static
         if (!dynamic && !pending)
         {
            pdata.staticResources.peek().add(res);
         }
         
         // need to init dynamic flag here to avoid NPE
         if (dynamic && res instanceof GenericWidget)
         {
            ((GenericWidget)res)._setDynamic(dynamic);
         }
         
         ((HandleResource) res).setInstantiatingProcedure(new handle(pdata.phandle));
      }
   }
   
   /**
    * Get the current external program (an application logic instance) being processed for 
    * scopeable or finalizable notifications.
    * 
    * @return   See above
    */
   public static Object getProcessedProcedure()
   {
      WorkArea wa = work.get();
      
      return wa.processedProcedure;
   }
   
   /**
    * Save the information about the program being called, to be used on the top-level scope 
    * start notification: at that time, the info about the callee (this, source- and target-
    * procedures and other info) will be resolved.
    * 
    * @param    superCall
    *           Flag indicating if the destination is in a super-procedure.
    * @param    target
    *           The target procedure, which was targeted by the caller.
    * @param    destination
    *           The procedure where the internal entry was resolved.
    * @param    iename
    *           The internal entry name. In case of triggers or external programs, use the 
    *           {@link #EXTERNAL_PROGRAM} and {@link #TRIGGER_PROCEDURE} constants.
    * @param    pname
    *           The external program name; will be non-null only when we are invoking an external
    *           program.
    * @param    function
    *           Flag indicating we are calling a user-defined function.
    * @param    persistentProc
    *           Flag indicating the external program was ran persistent.
    */
   public static void pushCalleeInfo(boolean superCall,
                                     Object  target,
                                     Object  destination,
                                     String  iename,
                                     String  pname,
                                     boolean function,
                                     boolean persistentProc)
   {
      CalleeInfoImpl info = new CalleeInfoImpl(superCall, target, destination, iename, pname, 
                                       function, persistentProc);

      work.get().calleeInfo.push(info);
   }
   
   /**
    * Pop the last callee info from the stack.
    */
   public static void popCalleeInfo()
   {
      work.get().calleeInfo.pop();
   }
   
   /**
    * Return the last callee info from the stack.
    * 
    * @return  {@link CalleeInfo} instance.
    */
   public static CalleeInfo peekCalleeInfo()
   {
      return work.get().calleeInfo.peek();
   }
   
   /**
    * Get the {@link Scopeable} instance for this context.
    * 
    * @return   See above.
    */
   static Scopeable getScopeable()
   {
      return work.get();
   }
   
   /**
    * Check if we are executing finalizable or scopeable notifications.
    * 
    * @return   See above.
    */
   static boolean inNotifications()
   {
      return work.get().processedProcedure != null;
   }
   
   /**
    * Set the external procedure beeing processed for scopeable or finalizable notifications.
    * 
    * @param    referent
    *           The external procedure instance.
    */
   static void setProcessedProcedure(Object referent)
   {
      WorkArea wa = work.get();
      
      wa.processedProcedure = referent;
   }
   
   /**
    * Notify all subscribers (to the given event) that the event has occurred.
    * <p>
    * All processing is done with silent error mode implicitly enabled.
    *
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event. 
    *           Must not be <code>null</code>.
    * @param    modes
    *           The mode for each parameter, encoded in a string. When <code>null</code>, 
    *           the parameter modes will not be checked against the procedure definition.
    * @param    param
    *           The parameters passed to the subscriber. May be <code>null</code> to represent no
    *           parameters. When <code>null</code>, it means that no parameters were specified for
    *           this call.
    */
   static void publish(character event, handle publisher, String modes, Object[] param)
   {
      ErrorManager.silentErrorEnable();
      try
      {
         if (publisher.isUnknown())
         {
            final String msg = "Could not evaluate publisher handle expression while trying to " +
                               "execute PUBLISH statement";
            ErrorManager.recordOrThrowError(6472, msg, false);
            return;
         }
         if (!isProcedure(publisher) || !publisher._isValid())
         {
            final String msg = 
               "Invalid or inappropriate publisher handle value given to PUBLISH statement";
            ErrorManager.recordOrThrowError(9072, msg, false);
            return;
         }
         
         param = (param == null ? new Object[0] : param);
         
         // Progress does not validate the event
         
         Object rpub = publisher.get();
         
         // walk through all the procedures which are subscribed to this event and publisher and 
         // invoke their method
         WorkArea wa = work.get();
         for (ProcedureData pp : wa.pMap.values())
         {
            String pname = getAbsoluteName(pp.phandle.get());
            String relName = getRelativeName(pp.phandle.get());

            // determine the subscriptions
            for (Subscription sub : pp.subscriptions)
            {
               // if the events match and the subscription is ANYWHERE or the publishers match,
               // then invoke the procedure
               if (sub.event.equals(event) &&
                  (sub.publisher == null || sub.publisher.equals(rpub)))
               {
                  // invoke the associated procedure. there are some rules here:
                  // - changes of INPUT-OUTPUT parameters are inherited from one call to another
                  // - OUTPUT parameters are always overridden
                  String iename = (sub.method.isUnknown() ? "" : sub.method.toStringMessage());
                  
                  // check if current procedure has a definition for target
                  String jname = SourceNameMapper.getMethodName(pname, iename, false);
                  if (jname == null)
                  {
                     final String msg = 
                        "PUBLISH statement failed. Procedure %s has no entry point for %s";
                     ErrorManager.recordOrThrowError(6480,
                                                     String.format(msg, relName, iename),
                                                     false);
                     continue;
                  }
                  
                  ExternalProgramWrapper extProg = new ExternalProgramWrapper(sub.subscriber);

                  // even if silent error mode is enabled by default, when the procedure is
                  // executed it will be disabled for the duration of the execution and re-enabled
                  // when the procedure ends. TransactionManager.push/popScope takes care of this.
                  ControlFlowOps.invokeIn(iename, new handle(extProg), param);
               }
            }
         }
      }
      finally
      {
         ErrorManager.silentErrorDisable();
      }
   }

   /**
    * Create a new subscription to the given event for a specified program. This program will 
    * receive a call-back when the event is published.
    *
    * @param    subscriber
    *           The program to notify when the specified event is published. Must not be 
    *           <code>null</code>.  The specified program must have a method that corresponds to 
    *           the given Progress 4GL internal procedure name (<code>method</code>).
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event. 
    *           If specified, the named event will only cause a notification if the event source 
    *           matches this handle. <code>null</code> represents the <code>ANYWHERE</code> option
    *           which means that the events will not be filtered by source.
    * @param    method
    *           The Progress 4GL internal procedure name which will be translated into a method 
    *           name and invoked on the <code>subscriber</code> instance. When <code>null</code>,
    *           it means that no method was specified for this call and the event name will be
    *           used as the method name.
    */
   static void subscribe(handle subscriber, character event, handle publisher, character method)
   {
      // validate subscriber
      if (subscriber.isUnknown())
      {
         final String msg = "Could not evaluate subscriber handle  expression while trying to " +
                            "execute SUBSCRIBE statement";
         ErrorManager.recordOrThrowError(6463, msg, false);
         return;
      }
      if (!isProcedure(subscriber) || !subscriber._isValid())
      {
         final String msg = "Invalid or inappropriate subscriber handle value given to " +
                            "SUBSCRIBE statement";
         ErrorManager.recordOrThrowError(6466, msg, false);
         return;
      }
      
      // validate publisher; if null, is OK, as it means ANYWHERE option
      if (publisher != null)
      {
         if (publisher.isUnknown())
         {
            final String msg = "Could not evaluate publisher handle  expression while trying " +
                               "to execute SUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6464, msg, false);
            return;
         }
         if (!isProcedure(publisher) || !publisher._isValid())
         {
            final String msg = "Invalid or inappropriate publisher handle value given to " +
                               "SUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6467, msg, false);
            return;
         }
      }
      // Progress does not validate event and method name
      
      Object rsub = subscriber.get();
      Object rpub = (publisher == null ? null : publisher.get());
      
      // 4GL doesn't raise an error if a subscription already exists, but doesn't replace it 
      // either
      ProcedureData pp = locateProcedure(rsub);
      boolean found = false;
      for (Subscription sub : pp.subscriptions)
      {
         if (sub.readOnly)
         {
            // read-only subscriptions are not checked.
            continue;
         }

         // get an exact match
         if (event.equals(sub.event) && 
             ((rpub == null && sub.publisher == null) || 
              (rpub != null && sub.publisher != null && rpub.equals(sub.publisher))))
         {
            found = true;
            break;
         }
      }

      // create only if there isn't an existing one
      if (!found)
      {
         Subscription sub = new Subscription(rsub, event, rpub, method);
         pp.subscriptions.add(sub);
      }
   }

   /**
    * Create a new subscription to the given event for any program. This program will receive a 
    * call-back when the event is published.
    *
    * @param    subscriber
    *           The program to notify when the specified event is published. Must not be 
    *           <code>null</code>.  The specified program must have a method that corresponds to 
    *           the given Progress 4GL internal procedure name (<code>method</code>).
    * @param    event
    *           The name of the event which is being published.
    * @param    method
    *           The Progress 4GL internal procedure name which will be translated into a method 
    *           name and invoked on the <code>subscriber</code> instance. When <code>null</code>,
    *           it means that no method was specified for this call and the event name will be
    *           used as the method name.
    */
   static void subscribeAnywhere(handle subscriber, character event, character method)
   {
      subscribe(subscriber, event, null, method);
   }
   
   /**
    * Delete a subscription to the given event for the specified program. This method is executed
    * in a {@link ErrorManager#silentErrorEnable()}/{@link ErrorManager#silentErrorDisable()}
    * bracket, as the <code>UNSUBSCRIBE</code> statement is executed with a default NO-ERROR 
    * clause.
    *
    * @param    subscriber
    *           The program whose subscription is to be deleted.  Must not be <code>null</code>.
    * @param    event
    *           The name of the event which is being published. If this is <code>null</code>, then
    *           subscriptions to all events will be deleted.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event.
    *           If specified, the subscription will only be deleted if the event source matches
    *           this handle.  <code>null</code> represents the <code>ALL</code> option which means
    *           that the events will not be filtered by source. When <code>null</code>, it means
    *           that the publisher was not set.
    */
   static void unsubscribe(handle subscriber, character event, handle publisher)
   {
      ErrorManager.silentErrorEnable();

      try
      {
         // validate subscriber
         if (subscriber.isUnknown())
         {
            final String msg = "Could not evaluate subscriber handle  expression while trying " +
                               "to execute UNSUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6475, msg, false);
            return;
         }
         if (!isProcedure(subscriber) || !subscriber._isValid())
         {
            final String msg = "Invalid or inappropriate subscriber handle value given to " +
                               "UNSUBSCRIBE statement";
            ErrorManager.recordOrThrowError(6477, msg, false);
            return;
         }
         
         // validate publisher; if null, is OK, as it means ALL option
         if (publisher != null)
         {
            if (publisher.isUnknown())
            {
               final String msg = "Could not evaluate publisher handle  expression while " +
                                  "trying to execute UNSUBSCRIBE statement";
               ErrorManager.recordOrThrowError(6476, msg, false);
               return;
            }
            if (!isProcedure(publisher) || !publisher._isValid())
            {
               final String msg = "Invalid or inappropriate publisher handle value given to " +
                                  "UNSUBSCRIBE statement";
               ErrorManager.recordOrThrowError(6478, msg, false);
               return;
            }
         }
         
         // Progress does not validate the event
         
         Object rsub = subscriber.get();
         Object rpub = (publisher == null ? null : publisher.get());
         
         ProcedureData pp = locateProcedure(rsub);
         boolean found = false;
         Iterator<Subscription> itr = pp.subscriptions.iterator();
         while (itr.hasNext())
         {
            Subscription sub = itr.next();

            // a special case: if a publisher is specified and ALL clause is used, than all 
            // ANYWERE subscriptions are marked read-only, for the event.
            if (rpub != null && event == null && sub.publisher == null)
            {
               sub.readOnly = true;
            }
            
            if (sub.readOnly)
            {
               // read-only subscriptions can't be removed
               continue;
            }
            
            // if removes all subscriptions considering:
            // - if the passed event is null, then ALL clause is in effect, else event names must
            // match
            // - the publisher is not matched as stated in the documentation. the correct approach
            // is that if the publisher is missing, then ANY publisher is matched
            if ((event == null || event.equals(sub.event)) &&
                ((rpub == null || rpub.equals(sub.publisher))))
            {
               // remove this subscription
               itr.remove();
               found = true;
            }
         }

         // if a specific event is used and no subscriptions found, then error
         if (!found && event != null)
         {
            final String msg = "UNSUBSCRIBE failed.  No subscriptions for %s";
            String sevt = (event.isUnknown() ? "" : event.toStringMessage());
            ErrorManager.recordOrThrowError(6479, String.format(msg, sevt), false);
            return;
         }
      }
      finally
      {
         ErrorManager.silentErrorDisable();
      }
   }

   /**
    * Delete a subscription to ALL events for the specified program.
    *
    * @param    subscriber
    *           The program whose subscription is to be deleted.  Must not be <code>null</code>.
    * @param    publisher
    *           The procedure handle representing the program that is the source of the event.
    *           If specified, the subscription will only be deleted if the event source matches
    *           this handle.  <code>null</code> represents the <code>ALL</code> option which means
    *           that the events will not be filtered by source.  When <code>null</code>, it means
    *           that the publisher was not set.
    */
   static void unsubscribeAll(handle subscriber, handle publisher)
   {
      unsubscribe(subscriber, null, publisher);
   }

   /**
    * Ensure all the context-local data has initialized (to be able to listen
    * for scope notifications).
    */
   static void load()
   {
      // needed to ensure the variable is loaded, so it listens for scope
      // notifications
      work.get();
   }
   
   /**
    * Check if the handle's referent is a procedure or not. This means that the value must be set 
    * and must be an instance of {@link ExternalProgramWrapper}. The enclosed referent may or may
    * not be registered with the {@link ProcedureManager}.
    *
    * @param    h
    *           handle object to check.
    *  
    * @return   See above.
    */
   public static boolean isProcedure(handle h)
   {
      return h != null && !h.isUnknown() && (h.getResource() instanceof ExternalProgramWrapper);
   }

   /**
    * Add the <code>h</code> handle as a persistent procedure. Used by 
    * {@link ControlFlowOps#invokePersistent(character, Object...)} and
    * {@link ControlFlowOps#invokePersistentSet(String, handle, Object...)}.
    * 
    * @param   h
    *          The procedure handle to be added to the list of persistent
    *          procedures.
    * @param   name
    *          The name of this procedure, used at the RUN statement.
    * @param   persistent
    *          Flag indicating if this is a persistent procedure.
    */
   static void addProcedure(handle h, String name, boolean persistent)
   {
      WorkArea wa = work.get();
      Object referent = h.get();
      ProcedureData pp = new ProcedureData(h, null);

      wa.pMap.put(referent, pp);

      if (persistent)
      {
         wa.persistentProcedures.add(referent);
      }
      wa.relativeNames.put(referent, name);
   }

   /**
    * Remove the given handle from the list of persistent procedures. Used by 
    * {@link ControlFlowOps#invokePersistent(character, Object...)} and
    * {@link ControlFlowOps#invokePersistentSet(String, handle, Object...)} in
    * cases when errors were encountered while invoking the procedure.
    * 
    * @param   h
    *          The procedure handle to be added to the list of persistent
    *          procedures.
    */
   static void removePersistentProcedure(handle h)
   {
      removeProcedureImpl(h, work.get().persistentProcedures);
   }
   
   /**
    * Gets the value of CURRENT-WINDOW attribute of specified procedure
    * @param   referent
    *          Procedure wrapper 
    * @return  handle.
    */
   public static handle getCurrentWindow(Object referent)
   {
      return new handle(locateProcedure(referent).currentWindow);
   }
   
   /**
    * Sets the value of CURRENT-WINDOW attribute of specified procedure
    * @param   referent
    *          Procedure wrapper
    * @param   winHandle 
    *          New value for the attribute.          
    */
   public static void setCurrentWindow(Object referent, handle winHandle)
   {
      if (winHandle == null || !winHandle._isValid() || winHandle.isUnknown())
      {
         ErrorManager.recordOrShowError(
            4083, 
            "Unable to assign UNKNOWN value to attribute CURRENT-WINDOW on PROCEDURE widget",
            false, false, false);
         
         return;
      }
      
      if (!(winHandle.getResource() instanceof WindowWidget))
      {
         String ms = "Unable to set attribute CURRENT-WINDOW of procedure widget";
         ErrorManager.recordOrShowError(4078, ms, false, false, false);
         
         return;
      }

      locateProcedure(referent).currentWindow.assign(winHandle);
   }
   
   /**
    * Implementation for the PROGRAM-NAME functions. Resolves the legacy top level block (external
    * program, function, internal procedure or trigger), at the given depth. The index starts with 
    * 0 for the currently executing top-level block.
    *  
    * @param    idx
    *           The stack entry index.
    * 
    * @return   The name of the stack entry or {@code null} if outside of stack.
    */
   static String getStackEntry(int idx)
   {
      WorkArea wa = work.get();
      
      if (idx < 0 || idx >= wa.thisProcedures.size())
      {
         return null;
      }
      
      
      Iterator<Object> procItr = wa.thisProcedures.iterator();
      Iterator<String> ieItr = wa.internalEntryStack.iterator();
      
      Object referent = null;
      String iename = null;
      
      for (int i = 0; i <= idx; i++)
      {
         referent = procItr.next();
         iename = ieItr.next();
      }
      
      final String fmt1 = "%s";
      final String fmt2 = "%s %s";
      
      String name = wa.relativeNames.get(referent);
      
      return (EXTERNAL_PROGRAM.equals(iename)) ? String.format(fmt1, name)
                                               : String.format(fmt2, iename, name);
   }

   /**
    * Get the relative name used to create the given external program instance.
    * 
    * @param    instance
    *           The external program instance.
    * 
    * @return   See above.
    */
   public static String getRelativeName(Object instance)
   {
      return work.get().relativeNames.get(instance);
   }

   /**
    * Push an internal entry to the stack.  The passed internal entry name is resolved and is
    * pushed as it was defined, not as it is being called.
    * <p>
    * Use {@link #EXTERNAL_PROGRAM} for the name of the root external procedure block and 
    * {@link #TRIGGER_PROCEDURE} for the root trigger block.
    * 
    * @param    instance
    *           The external program instance.
    * @param    iename
    *           The legacy internal-entry name.
    * @param    function
    *           Flag indicating if the internal entry is a user-defined function.
    */
   private static void pushInternalEntry(Object instance, String iename, boolean function)
   {
      WorkArea wa = work.get();
      
      if (!(iename.equals(EXTERNAL_PROGRAM) || iename.equals(TRIGGER_PROCEDURE)))
      {
         // the internal entry names are reported as they are defined, not as they are called.
         // thus, we need to determine the definition and get it from there
         
         String pname = getAbsoluteName(instance);
         InternalEntry ie = SourceNameMapper.getInternalEntry(pname, iename, function);
         
         iename = (ie == null ? iename : ie.pname);
      }

      wa.internalEntryStack.push(iename);
   }
   
   /**
    * Get the absolute name of the external program associated with the given java instance.
    * <p>
    * This is used internally by P2J, as we need absolute names when resolving a program's
    * signature, events, etc.
    * <p>
    * For the name associated with the procedure's NAME and FILE-NAME attributes, and returned by
    * the PROGRAM-NAME function or used in 4GL error messages, see {@link #getRelativeName}, which
    * returns the real name used when the external procedure was instantiated first.
    * 
    * @param    referent
    *           The referent for which the name is needed.
    *           
    * @return   The absolute name.
    */
   static String getAbsoluteName(Object referent)
   {
      String jname = referent.getClass().getName();
      String pname = SourceNameMapper.convertJavaProg(jname);
      return pname;
   }

   /**
    * Check if the given referent has been (or is still) registered with the
    * {@link ProcedureManager} in the {@link WorkArea#pMap procedure} map.
    * 
    * @param     referent
    *            The referent to be checked if is still loaded.
    * 
    * @return    <code>true</code> if the referent was found.
    */
   static boolean hasReferent(Object referent)
   {
      return work.get().pMap.containsKey(referent);
   }
   
   /**
    * Delete the given procedure handle (i.e. execute the DELETE OBJECT or DELETE PROCEDURE 
    * handle).
    * 
    * @param    referent
    *           The persistent procedure instance to be deleted.
    * 
    * @return   <code>true</code> if this procedure handle can be deleted.
    */
   static boolean delete(Object referent)
   {
      // handle is always valid
      WorkArea wa = work.get();

      ProcedureData pdata = locateProcedure(referent);
      
      if (_isPersistent(referent) && systemHandle(referent))
      {
         // if this is a persistent system handle, the actual delete is postponed until we
         // are finished.
         wa.deleteOnExit.put(referent, true);
         
         return false;
      }
      
      wa.persistentProcedures.remove(referent);
      wa.sessionSuperProcs.remove(referent);
      
      for (ProcedureData pp : wa.pMap.values())
      {
         pp.superProcedures.remove(referent);
      }

      wa.cleanup(referent);
      
      try
      {
         wa.processedProcedure = referent;
         
         // cleanup after the procedure
         for (Finalizable target : pdata.finalizables)
         {
            target.deleted();
         }
         
         for (Scopeable target : pdata.scopeables)
         {
            target.scopeDeleted();
         }

         // as the procedure no longer exists, delete the static resources
         deleteResources(pdata.staticResources.pop());
      }
      finally
      {
         wa.processedProcedure = null;
      }

      // delete the widget pools for this external program
      WidgetPool.deleteWidgetPools(referent);
      
      return true;
   }

   /**
    * Using the name of a function defined IN handle and a procedure handle 
    * to which this function definition belongs, resolve the target handle,
    * to be able to invoke it.
    * 
    * @param    phandle
    *           The procedure handle where the function was defined.
    * @param    fname
    *           The function's legacy name.
    *           
    * @return   The procedure handle where to search for the function's body.
    */
   static handle getHandleForFunction(handle phandle, String fname)
   {
      Object referent = phandle.get();
      ProcedureData pp = locateProcedure(referent);
   
      handle h = (handle) pp.inHandleFunctions.get(fname).resolve();
      return h == null ? new handle() : h;
   }

   /**
    * For procedures or functions defined IN SUPER, this computes the target procedure handle 
    * where the procedure's or function's body resides.
    * <p>
    * In case when no super procedure is found, it will throw proper error messages and return
    * <code>null</code>. 
    * 
    * @param    iename
    *           Internal-entry's name (this may be a procedure or function).
    * @param    function
    *           <code>true</code> if this is a function.
    *           
    * @return   A procedure handle where to search for the internal-entry's body or
    *           <code>null</code> of none found.
    */
   static handle getSuperHandle(String iename, boolean function)
   {
      WorkArea wa = work.get();
      
      handle hthis           = thisProcedure();
      handle htarget         = targetProcedure();
      Object refThis         = hthis.get();
      Object refTarget       = htarget.get();
      ProcedureData ppThis   = locateProcedure(refThis);
      ProcedureData ppTarget = locateProcedure(refTarget);

      boolean searchSelf = ppThis.isSearchSelf();
      handle psuper = null;

      @SuppressWarnings("unchecked")
      Collection<Object>[] search = new Collection[2];
      
      if (searchSelf)
      {
         search[0] = ppThis.superProcedures.keySet();
         search[1] = wa.sessionSuperProcs.keySet();
      }
      else
      {
         Collection<Object> targetProcs = ppTarget.superProcedures.keySet();
         search[0] = new ArrayDeque<>(targetProcs);
         search[1] = new ArrayDeque<>(wa.sessionSuperProcs.keySet());
         
         boolean inTarget = search[0].contains(refThis);
         boolean inSess   = search[1].contains(refThis);
         
         if (inTarget || inSess)
         {
            // remove referents until this-procedure is found in the list
            prepareSearches(search, refThis);
         }
         
         if (!(inTarget || inSess) || (inSess && search[1].isEmpty() && !targetProcs.isEmpty()))
         {
            if (!function)
            {
               // in some cases when invoking internal procedures, it can end up in an infinite 
               // recursion
   
               search[0] = new ArrayDeque<>(targetProcs);
               search[1] = new ArrayDeque<>(wa.sessionSuperProcs.keySet());
               
               // step 1. find the first already-invoked proc part of target's super-procs
               Object ref = null;
               for (CalleeInfoImpl info : wa.calleeInfo)
               {
                  if (targetProcs.contains(info.destination))
                  {
                     ref = info.destination;
                     break;
                  }

                  if (!info.superCall && refTarget != info.destination)
                  {
                     break;
                  }
               }

               if (ref != null)
               {
                  // step 2. remove all until ref is found, from all searches
                  prepareSearches(search, ref);
               }
               else
               {
                  search[0] = null;
                  search[1] = null;
               }
            }
            else
            {
               // not found, clear the lists
               search[0] = null;
               search[1] = null;
            }
         }
      }
      
      for (int i = 0; i < search.length && psuper == null; i++)
      {
         if (search[i] == null || search[i].isEmpty())
         {
            continue;
         }
         
         // search is done backwards (last set super-procedure is searched first)
         Deque<Object> queue = new ArrayDeque<>(search[i]);
         while (!queue.isEmpty())
         {
            Object referent = queue.removeLast();
            ProcedureData pp = locateProcedure(referent);

            if (pp == null || !pp.phandle._isValid())
            {
               // deleted or invalid procedures are ignored
               continue;
            }

            String spname = getAbsoluteName(pp.phandle.get());
            String jname = SourceNameMapper.getMethodName(spname, iename, function);
            if (jname != null)
            {
               // found it
               psuper = pp.phandle;
               break;
            }
         }
      }
      
      if (psuper == null)
      {
         String err = null;
         int errId = -1;
         if (function)
         {
            final String errMsg = 
               "SUPER version of user-defined function '%s' invoked but could not be found";
            errId = 6454;
            err = String.format(errMsg, iename);

            ErrorManager.recordOrShowError(errId, err, false, false, false);
         }
         else
         {
            String pname = getRelativeName(searchSelf ? refThis : refTarget);

            errId = 6439;
            if (searchSelf)
            {
               final String errMsg = 
                  "Procedure %s %s has no SUPER procedure with internal procedure %s";
               err = String.format(errMsg, iename, pname, iename);
            }
            else
            {
               final String errMsg = 
                 "Procedure %s has no SUPER procedure with internal procedure %s";
               err = String.format(errMsg, pname, iename);
            }

            ErrorManager.recordOrThrowError(errId, err, false);
         }
      }
      
      return psuper;
   }
   
   /**
    * Get the THIS-PROCEDURE referent.
    * 
    * @return   See above.
    */
   static Object _thisProcedure()
   {
      WorkArea wa = work.get();
      Object referent = wa.thisProcedures.peek();
      
      return referent;
   }

   /**
    * Check if the given referent is for a persistent procedure. This 
    * implements the PERSISTENT attribute for a procedure handle.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *
    * @return   <code>true</code> if the referent is a persistent procedure.
    */
   static logical isPersistent(Object referent)
   {
      return new logical(_isPersistent(referent));
   }

   /**
    * Check if the given referent is for a persistent procedure. This 
    * implements the PERSISTENT attribute for a procedure handle.
    * 
    * @param    referent
    *           An instance of a converted external procedure class.
    *
    * @return   <code>true</code> if the referent is a persistent procedure.
    */
   public static boolean _isPersistent(Object referent)
   {
      return work.get().persistentProcedures.contains(referent);
   }
   
   /**
    * Check if this is a "delete-on-exit" procedure.
    * 
    * @param    referent
    *           An instance of a converted external procedure class.
    *           
    * @return   <code>true</code> if the referent is a "delete-on-exit" procedure.
    */
   static boolean _isDelayedDelete(Object referent)
   {
      Boolean delete = work.get().deleteOnExit.get(referent);
      
      return delete != null && delete.booleanValue();
   }
   
   /**
    * Check if the given referent is the top-level procedure (the startup procedure).
    * 
    * @param    referent
    *           The instance to be verified.
    *           
    * @return   <code>true</code> if the referent is the startup procedure.
    */
   static boolean isTopLevel(Object referent)
   {
      Deque<Object> procs = work.get().thisProcedures;
      return procs.getLast() == referent;
   }

   /**
    * Check if the given handle is a proxy procedure.
    * 
    * @param    h
    *           The handle to check.
    *           
    * @return   <code>true</code> if the handle is a proxy procedure.
    */
   static boolean isProxy(handle h)
   {
      return h != null && h._isValid() && (h.getResource() instanceof ProxyProcedureWrapper);
   }

   /**
    * Get the <code>SUPER-PROCEDURES</code> attribute of the specified 
    * external procedure instance.
    *
    * @param    referent
    *           An instance of an converted external procedure class.
    * 
    * @return   See above.
    */
   static character superProcedures(Object referent)
   {
      WorkArea wa = work.get();
      return superProceduresImpl(wa.getSuperProcedures(referent).values());
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h)
   {
      return addSuperProcedure(referent, h, (SearchMode) null);
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent,
    * using the explicit {@link SearchMode mode}.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The {@link SearchMode} to be set for the <code>h</code> 
    *           handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h,SearchMode mode)
   {
      return addSuperProcedureImpl(h, mode, referent);
   }
   
   /**
    * Add the given handle as a super-procedure for the specified referent,
    * using the search-mode set as a string (one of the 
    * <code>SEARCH-TARGET</code> or <code>SEARCH-SELF</code> string literals).
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The handle to be added as a super-procedure.
    * @param    mode
    *           The search-mode to be set for the <code>h</code> handle.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical addSuperProcedure(Object referent, handle h, int mode)
   {
      return addSuperProcedure(referent, h, getSearchMode(mode));
   }

   /**
    * Remove the given handle from the specified referent's list of 
    * super-procedures.
    *  
    * @param    referent
    *           An instance of an converted external procedure class.
    * @param    h
    *           The procedure handle to be removed.
    *           
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   static logical removeSuperProcedure(Object referent, handle h)
   {
      if (!h._isValid() || !isProcedure(h))
      {
         return new logical(false);
      }

      WorkArea wa = work.get();
      removeProcedureImpl(h, wa.getSuperProcedures(referent).keySet());
      
      return new logical(true);
   }
   
   /**
    * Get the <code>NEXT-SIBLING</code> attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static handle getNextSibling(Object referent)
   {
      WorkArea wa = work.get();
      
      Object prev = null;
      Object next = null;
      for (Object pp : wa.persistentProcedures)
      {
         if (prev != null && prev == referent)
         {
            next = pp;
            break;
         }

         prev = pp;
      }
      
      if (next == null)
      {
         return new handle();
      }

      return locateProcedure(next).phandle;
   }
   
   /**
    * Get the <code>PREV-SIBLING</code> attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static handle getPrevSibling(Object referent)
   {
      WorkArea wa = work.get();
      
      Object prev = null;
      for (Object pp : wa.persistentProcedures)
      {
         if (pp == referent)
         {
            break;
         }
         prev = pp;
      }
      
      if (prev == null)
      {
         return new handle();
      }

      return locateProcedure(prev).phandle;
   }

   static void setPrivateData(Object referent, character data)
   {
      ProcedureData pp = locateProcedure(referent);
      pp.privateData = new character(data);
   }

   /**
    * Get the <code>PRIVATE-DATA</code> attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static character getPrivateData(Object referent)
   {
      ProcedureData pp = locateProcedure(referent);

      return pp.privateData == null ? new character() 
                                    : new character(pp.privateData);
   }
   
   /**
    * Get the value of the ADM-DATA attribute of the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    * 
    * @return   See above.
    */
   static character getADMData(Object referent)
   {
      ProcedureData pp = locateProcedure(referent);
      return new character(pp.admData);
   }

   /**
    * Set the value of the ADM-DATA attribute for the specified referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *
    * @param    value
    *           The new value.
    */
   static void setADMData(Object referent, String value)
   {
      ProcedureData pp = locateProcedure(referent);
      pp.admData.assign(value);
   }
   
   /**
    * Gets the the unique ID number associated to the referent. Not the 
    * same as the handle value itself.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *
    * @return   The integer number of the children.
    */
   static integer getUniqueID(Object referent)
   {
      ProcedureData pp = locateProcedure(referent);
      return new integer(pp.uniqueID);
   }

   /**
    * Get a handle list of super-procedures for the given referent.
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static List<handle> getSuperProcedures(Object referent)
   {
      WorkArea wa = work.get();
      
      List<handle> superProcs = new LinkedList<handle>();
      // need to emit them in reverse order (they are kept in LIFO)
      for (ProcedureData pp : wa.getSuperProcedures(referent).values())
      {
         superProcs.add(0, pp.phandle);
      }
      
      return superProcs;
   }
   
   /**
    * Get a handle list of super-procedures for the session handle.
    *           
    * @return   See above.
    */
   static List<handle> getSuperProcedures()
   {
      return getSuperProcedures(null);
   }
   
   /**
    * Check if the given referent is for a system handle (i.e. for one of the
    * THIS-PROCEDURE, SOURCE-PROCEDURE or TARGET-PROCEDURE handles).
    * 
    * @param    referent
    *           An instance of an converted external procedure class.
    *           
    * @return   See above.
    */
   static boolean systemHandle(Object referent)
   {
      WorkArea wa = work.get();
      return wa.thisProcedures.contains(referent) ||
             wa.targetProcedures.contains(referent) ||
             wa.sourceProcedures.contains(referent);
   }
   
   /**
    * Register the given {@link Scopeable} to be processed when the persistent procedure gets
    * deleted.
    * 
    * @param    referent
    *           The external program.
    * @param    target
    *           The {@link Scopeable} instance which needs delayed notification.
    */
   static void addScopeable(Object referent, Scopeable target)
   {
      ProcedureData pdata = locateProcedure(referent);
      pdata.scopeables.add(target);
   }
   
   /**
    * Register the given {@link Finalizable} to be processed when the persistent procedure gets
    * deleted.
    * 
    * @param    referent
    *           The external program.
    * @param    target
    *           The {@link Finalizable} instance which needs delayed notification.
    */
   static void addFinalizable(Object referent, Finalizable target)
   {
      ProcedureData pdata = locateProcedure(referent);
      pdata.finalizables.add(target);
   }
   
   /**
    * Gets the INSTANTIATING-PROCEDURE associated to the referent. 
    * 
    * @param    referent
    *           The external program.
    * 
    * @return   See above.
    */
   static handle getInstantiatingProcedure(Object referent)
   {
      ProcedureData pdata = locateProcedure(referent);
      return new handle(pdata.instProc);
   }
   
   /**
    * Resolve the instantiating procedure from the given handle: if the handle contains a valid
    * procedure, return a copy of it; else, return an unknown handle.
    * 
    * @param    hproc
    *           The handle instance representing an instantiating procedure.
    * 
    * @return   See above.
    */
   static handle resolveInstantiatingProcedure(handle hproc)
   {
      return hproc._isValid() ? new handle(hproc) : new handle();
   }
   
   /**
    * Resolve the closest business logic method in the stack trace.
    * 
    * @param    referent
    *           The external program to which the method must belong.
    * @param    function
    *           Flag indicating if a function or internal procedure will be searched.
    * @param    externalProc
    *           Flag indicating if we are looking for an external procedure.
    *          
    * @return   The legacy method name.
    */
   static String resolveClosestMethod(Object referent, boolean function, boolean externalProc)
   {
      // get a helper to process the stack
      Throwable stack = new Throwable();
      
      // obtain the stack trace
      stack.fillInStackTrace();
      StackTraceElement[] trace = stack.getStackTrace();
      
      String rClsName   = referent.getClass().getName();
      String methodName = null;
      String legacyName = null;
      
      String bmClass = BlockManager.class.getName();
      String procApi = (externalProc ? "internalProcedure" : "externalProcedure");
      String funcApi = "function";

      // walk the stack trace until we find a method defined by the referent's class
      boolean foundBlockManager = false;
      for (int i = 1; i < trace.length; i++)
      {
         String cname = trace[i - 1].getClassName();
         String mname = trace[i - 1].getMethodName();
         
         if (cname.equals(bmClass) && 
             ((function && mname.toLowerCase().endsWith(funcApi)) || 
              (!function && mname.equals(procApi))))
         {
            foundBlockManager = true;
         }
         else if (foundBlockManager)
         {
            methodName = mname;
            break;
         }
      }
      
      if (methodName != null)
      {
         if ("execute".equals(methodName))
         {
            // this is a special 4GL string.
            legacyName = EXTERNAL_PROGRAM;
         }
         else
         {
            // resolve the legacy name of this method
            legacyName = SourceNameMapper.convertJavaMethod(rClsName, methodName, function);
            
            if (legacyName == null)
            {
               // assume the java name as legacy name
               legacyName = methodName;
            }
         }
      }
      
      if (legacyName == null)
      {
         final String msg = "Could not resolve closest legacy method for class %s";
         throw new RuntimeException(String.format(msg, rClsName) , stack);
      }
      
      return legacyName;
   }
   
   /**
    * Find the root enclosing instance of the specified inner class instance.
    * It gets computed using the <code>this$</code> fields inserted by the 
    * compiler; it returns the instance refered by the <code>this$0</code> 
    * fields, when reached.
    * 
    * @param    instance
    *           The inner class instance.
    *           
    * @return   The root instance.
    */
   static Object findRootEnclosingInstance(Object instance)
   {
      Field[] fields = instance.getClass().getDeclaredFields();
      for (int i = 0; i < fields.length; i++)
      {
         Field f = fields[i];
         String fname = f.getName();
         if (fname.startsWith("this$"))
         {
            try
            {
               f.setAccessible(true);
               Object outer = f.get(instance);
               
               return 
                  fname.equals("this$0") ? outer 
                                         : findRootEnclosingInstance(outer);
            }
            catch (Exception e)
            {
               e.printStackTrace();
               return null;
            }
         }
      }
      
      return null;
   }

   /**
    * Implementation for the SUPER-PROCEDURES attribute.
    *  
    * @param    superProcs
    *           A list of {@link ProcedureData} objects, in the order in 
    *           which they were added.
    *           
    * @return   The string representation of each super-procedure, separated
    *           by a comma. The last added super-procedure must be first in
    *           this list, the second-to-last must be the second in the list,
    *           etc.
    */
   private static character superProceduresImpl(Collection<ProcedureData> superProcs)
   {
      character sp = new character("");
      boolean first = true;
      ArrayDeque<ProcedureData> queue = new ArrayDeque<ProcedureData>();
      queue.addAll(superProcs);
      // need to emit them in a LIFO order
      while (!queue.isEmpty())
      {
         ProcedureData pp = queue.removeLast();
         long hashCode = handle.resourceId(pp.phandle.getResource());
         sp.assign(TextOps.concat(sp, (!first ? "," : ""), hashCode));
         first = false;
      }
      
      return sp;
   }
   
   /**
    * The implementation of the ADD-SUPER-PROCEDURE method. When the
    * <code>referent</code> is null, this call is for the SESSION handle.
    * 
    * @param    h
    *           A handle to be added as super-procedure.
    * @param    mode
    *           The {@link SearchMode} for the passed handle.
    * @param    referent
    *           When <code>null</code> add to the SESSION:SUPER-PROCEDURES
    *           list. Else, add to the procedure associated with this 
    *           referent.
    * 
    * @return   <code>true</code> if <code>h</code> is a valid procedure
    *           handle.
    */
   private static logical addSuperProcedureImpl(handle h, SearchMode mode, Object referent)
   {
      if (!h._isValid() || !isProcedure(h))
      {
         return new logical(false);
      }
      
      WorkArea wa = work.get();
      Map<Object, ProcedureData> superProcs = wa.getSuperProcedures(referent);
      ProcedureData pData = locateProcedure(h.get());

      // the super-procedure to be added to the chain will always exist (either as the current
      // this-procedure or as a persistent procedure). the remove call is needed just to eliminate
      // it from the referent's super procedure chain, if was previously added 
      removeProcedureImpl(h, superProcs.keySet());

      
      if (!(mode == null || pData.searchMode == null || pData.searchMode.equals(mode)))
      {
         // if the mode is changed, show warning
         int errNum = 8914;
         String msg = "**Changing proc-search for procedure %s from  '%s' to '%s'";

         String progName = getRelativeName(referent);
         String oldMode  = pData.searchMode.toString();
         String newMode  = mode.toString();
         
         msg = String.format(msg, progName, oldMode, newMode);

         // this is just a warning, the mode will be changed
         ErrorManager.recordOrShowError(errNum, msg, false, false, false);
      }

      pData.searchMode = mode;
      superProcs.put(h.get(), pData);

      return new logical(true);
   }

   /**
    * Removes the given procedure handle from the given procedure instances list.
    * 
    * @param    h
    *           The procedure handle to be removed.
    * @param    ppList
    *           The collection from which the procedure must be removed.
    */
   private static void removeProcedureImpl(handle h, Collection<Object> ppList)
   {
      WorkArea wa = work.get();
      
      Object hreferent = h.get();
      
      Iterator<Object> iter = ppList.iterator();
      while (iter.hasNext())
      {
         Object referent = iter.next();
         if (hreferent == referent)
         {
            iter.remove();
         }
      }

      wa.cleanup(hreferent);
   }
   
   /**
    * Get the {@link SearchMode} constant associated with the given mode, used
    * by ADD-SUPER-PROCEDURE calls. If the mode is not one of the SEARCH-SELF 
    * or SEARCH-TARGET values, or an {@link IllegalArgumentException} will be 
    * thrown.
    * <p>
    * The given <code>mode</code> is not compatible with the mode returned by
    * {@link SearchMode#toString()}.
    * 
    * @param    mode
    *           The string representation of the search-mode.
    *           
    * @return   For SEARCH-TARGET value, {@link SearchMode#SEARCH_TARGET} 
    *           will be returned. For SEARCH-SELF value, 
    *           {@link SearchMode#SEARCH_SELF} will be returned.
    *
    * @throws   IllegalArgumentException
    *           If the mode is not of a known value.
    */
   private static SearchMode getSearchMode(int mode)
   {
      if (mode == SEARCH_SELF)
         return SearchMode.SEARCH_SELF;
      
      if (mode == SEARCH_TARGET)
         return SearchMode.SEARCH_TARGET;
      
      // if not a known mode, default to "search mode not specified"
      return null;
   }
   
   /**
    * Locate the {@link ProcedureData} instance associated with the given external program
    * instance.
    * <p>
    * This will find any persistent procedure or any procedure active on the stack.
    * 
    * @param    referent
    *           The external program instance.
    * 
    * @return   See above.
    */
   private static ProcedureData locateProcedure(Object referent)
   {
      return work.get().pMap.get(referent);
   }
   
   /**
    * Delete the given resources, as long as they are still valid.
    * 
    * @param    resources
    *           The resources to be deleted.
    */
   private static void deleteResources(Set<WrappedResource> resources)
   {
      for (WrappedResource target : resources)
      {
         if (target.valid())
         {
            ((Deletable) target).delete();
         }
      }
   }
   
   /**
    * Remove elements from the end of each queue until the specified referent is found.
    * 
    * @param   queues
    *          The queues which need to be prepared.
    * @param   ref
    *          The referent on which removing will stop.  This will be also removed, if found.
    */
   private static void prepareSearches(Collection<?>[] queues, Object ref)
   {
      l1: for (int i = 0; i < queues.length; i++)
      {
         Deque<Object> queue = new ArrayDeque<>(queues[i]);
         while (!queue.isEmpty())
         {
            Object referent = queue.removeLast();
            
            queues[i].remove(referent);
            
            if (referent == ref)
            {
               break l1;
            }
         }
      }
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   implements Scopeable
   {
      /**
       * The list of all registered persistent procedures, via RUN PERSISTENT
       * calls.
       */
      private Deque<Object> persistentProcedures = new ArrayDeque<Object>();
      
      /**
       * A map containing the list of super-procedures, for the session,
       * by each instance. 
       */
      private Map<Object, ProcedureData> sessionSuperProcs = 
         new LinkedHashMap<Object, ProcedureData>();
      
      /** 
       * This map holds the attributes for all registered external procedures, by instance.
       * It includes all persistent procedures and all procedures active on the stack.
       */
      private Map<Object, ProcedureData> pMap = new HashMap<Object, ProcedureData>();
      
      /** Stack of target-procedures, by instance. */
      private Deque<Object> targetProcedures = new ArrayDeque<Object>();

      /** Stack of source-procedures, by instance. */
      private Deque<Object> sourceProcedures = new ArrayDeque<Object>();
      
      /** Stack of this-procedures, by instance. */
      private Deque<Object> thisProcedures = new ArrayDeque<Object>();
      
      /**
       * Stack of internal entry names.  Evolves together with the {@link #thisProcedures} stack. 
       */
      private Deque<String> internalEntryStack = new ArrayDeque<>();
      
      /** Resources pending to registered with THIS-PROCEDURE when the next block is entered. */
      private Set<WrappedResource> pendingResources = new HashSet<>();
      
      /** Map of relative names (i.e. used at the RUN statement) for each external program. */
      private Map<Object, String> relativeNames = new HashMap<Object, String>();
      
      /**
       * Indicates the current external program when a scope or finalizable notification occurs.
       */
      private Object processedProcedure = null;

      /** A stack containing the info about the next called program or internal entry. */
      private Deque<CalleeInfoImpl> calleeInfo = new ArrayDeque<>();
      
      /** Stack with flags indicating if this procedure should be deleted on exit. */
      private Map<Object, Boolean> deleteOnExit = new HashMap<Object, Boolean>();
      
      /**
       * Handle scope notifications, to maintain proper source, target, and
       * this-procedure references.
       */
      @Override
      public void scopeStart()
      {
         BlockType bt = TransactionManager.getBlockType();

         if (bt.equals(BlockType.EXTERNAL_PROC) ||
             bt.equals(BlockType.INTERNAL_PROC) ||
             bt.equals(BlockType.FUNCTION)      ||
             bt.equals(BlockType.TRIGGER))
         {
            // push always before pushing new THIS-PROCEDURE
            CalleeInfoImpl info = calleeInfo.peek();
            info.push();
            
            ProcedureData pdata = pMap.get(info.destination);

            // setting INSTANTIATING-PROCEDURE for external procedure block can be done only 
            // after ProcedureData is created
            if (bt.equals(BlockType.EXTERNAL_PROC))
            {
               int nestingLevel = TransactionManager.getNestingLevel();
               if (AppServerManager.isRemote() ? nestingLevel > 3 : nestingLevel > 2)
               {
                  // if this is not the entry point, set the INSTANTIATING-PROCEDURE as the
                  // current THIS-PROCEDURE
                  pdata.instProc.assign(thisProcedure());
               }
            }
            
            thisProcedures.push(info.destination);

            WidgetPool.scopeStart();

            pdata.staticResources.push(new LinkedHashSet<WrappedResource>());
         }

         if (!TransactionManager.isGlobalBlock())
         {
            ProcedureData pdata = locateProcedure(_thisProcedure());
   
            // process the newly created resources
            for (WrappedResource res : pendingResources)
            {
               if (res instanceof HandleResource)
               {
                  ((HandleResource) res).setInstantiatingProcedure(new handle(pdata.phandle));
               }
            }

            // collect the static resources in the correct scope
            pdata.staticResources.peek().addAll(pendingResources);
            pendingResources.clear();
         }
      }

      /**
       * Handle scope notifications, to maintain proper source, target, and
       * this-procedure references.
       */
      @Override
      public void scopeFinished()
      {
         BlockType bt = TransactionManager.getBlockType();
         if (bt.equals(BlockType.EXTERNAL_PROC) ||
             bt.equals(BlockType.INTERNAL_PROC) ||
             bt.equals(BlockType.FUNCTION)      ||
             bt.equals(BlockType.TRIGGER))
         {
            // pop this-procedure from the stack and cleanup
            Object referent = thisProcedures.pop();
            ProcedureData pdata = locateProcedure(referent);

            CalleeInfoImpl info = calleeInfo.peek();
            info.pop();
            
            boolean extProc = bt.equals(BlockType.EXTERNAL_PROC);
            boolean persistent = _isPersistent(referent);
            boolean postponedDelete = false;
            
            if (extProc)
            {
               // check if we have a postponed delete
               Boolean delete = deleteOnExit.remove(referent);
               
               if (delete != null && delete.booleanValue())
               {
                  // this is no longer a persistent procedure
                  persistentProcedures.remove(referent);
                  persistent = false;
                  
                  // force the delete
                  delete(referent);
                  
                  postponedDelete = true;
               }
            }

            cleanup(referent);
            
            // notify the widget-pool management a procedure has finished; this needs to be done
            // here to ensure that the correct THIS-PROCEDURE handle is available
            if (!postponedDelete && !(extProc && persistent))
            {
               // delete the static resources created in an internal proc/function/trigger or
               // if the external program is not persistent
               // bypass this if there was a postponed delete, as the resources were deleted by 
               // the delete(referent) call 
               deleteResources(pdata.staticResources.pop());
            }

            WidgetPool.scopeFinished(referent, extProc, persistent);
         }
      }
      
      /**
       * Provides a notification that an external scope is about to be deleted.
       * <p>
       * This implementation is a no-op.
       */
      @Override
      public void scopeDeleted()
      {
         // no-op
      }
      
      /**
       * Get the map of session's super-procedures.
       */
      public Map<Object, ProcedureData> getSessionSuperProcedures()
      {
         return getSuperProcedures(null);
      }

      /**
       * Get the map of the super-procedures for the given referent.
       * 
       * @param    referent
       *           The external program instance for which the super-procedures
       *           are needed.
       */
      public Map<Object, ProcedureData> getSuperProcedures(Object referent)
      {
         return referent == null ? sessionSuperProcs 
                                 : pMap.get(referent).superProcedures;
      }
      
      /**
       * Attempt cleanup of the given external program instance. If is no
       * longer used as a persistent procedure, super procedure, or as a 
       * system handle, the referent will be removed from the {@link #pMap}
       * map.
       * 
       * @param    referent
       *           The external program instance for which needs to be cleaned
       *           up.
       */
      public void cleanup(Object referent)
      {
         // if no longer used, remove from global map
         if (!thisProcedures.contains(referent)                   &&
             !targetProcedures.contains(referent)                 &&
             !sourceProcedures.contains(referent)                 &&
             !persistentProcedures.contains(referent)             &&
             !sessionSuperProcs.containsKey(referent)             &&
             !superProcForExtProc(referent, persistentProcedures) &&
             !superProcForExtProc(referent, sessionSuperProcs.keySet()))
         {
            pMap.remove(referent);
            relativeNames.remove(referent);
         }
      }
      
      /**
       * Check if the given referent is a super-procedure of one of the
       * procedures in the given list.
       * 
       * @param    referent
       *           The external program instance.
       * @param    progs
       *           The list of external programs for which their 
       *           super-procedures list will be searched for the referent.
       *            
       * @return   <code>true</code> if it was found as a super-procedure for
       *           one of the given procedures.
       */
      private boolean superProcForExtProc(Object referent, 
                                          Collection<Object> progs)
      {
         for (Object prog : progs)
         {
            ProcedureData pp = pMap.get(prog);
            for (Object sreferent : pp.superProcedures.keySet())
            {
               if (sreferent == referent)
               {
                  return true;
               }
            }
         }
         
         return false;
      }
   }

   /**
    * Contains data about a subscription made by a certain procedure.
    */
   private static class Subscription
   {
      /** The subscriber. */
      private final Object subscriber;
      
      /** The subscribed event. */
      private final character event = new character();
      
      /** The method to be invoked. May be <code>null</code>. */
      private final character method;
      
      /** The publisher to match. <code>null</code> represents any source. */
      private final Object publisher;
      
      /** Flag indicating that this subscription can't be removed. */
      public boolean readOnly = false;
      
      /**
       * Build a new subscription with the specified data.
       * 
       * @param    subscriber
       *           The subscriber procedure.
       * @param    event
       *           The event to subscribe.
       * @param    publisher
       *           The event's publisher. May be <code>null</code>.
       * @param    method
       *           The method to be invoked. When <code>null</code>, the event name will be used
       *           as the method name.
       */
      public Subscription(Object subscriber, character event, Object publisher, character method)
      {
         this.subscriber = subscriber;
         this.publisher = publisher;
         this.event.assign(event);
         this.method = new character(method == null ? event : method);
      }
   }
   
   /**
    * A container with various data about the next called program/internal-entry, needed by
    * {@link WorkArea#scopeStart()}.
    */
   private static class CalleeInfoImpl 
   implements CalleeInfo
   {
      /** Flag indicating if the destination is in a super-procedure. */
      private final boolean superCall;
      
      /** The target procedure, which was targeted by the caller. */
      private final Object target;
      
      /** The procedure where the internal entry was resolved. */
      private final Object destination;
      
      /** 
       * The internal entry name. In case of triggers or external programs, use the 
       * {@link #EXTERNAL_PROGRAM} and {@link #TRIGGER_PROCEDURE} constants.
       */
      private final String iename;
      
      /** 
       * The external program name; will be non-null only when we are invoking an external 
       * program.
       */
      private final String pname;
      
      /** Flag indicating we are calling a user-defined function. */
      private final boolean function;
      
      /** Flag indicating the external program was ran persistent. */
      private final boolean persistentProc;
      
      /**
       * Create a new instance.
       * 
       * @param    superCall
       *           Flag indicating if the destination is in a super-procedure.
       * @param    target
       *           The target procedure, which was targeted by the caller.
       * @param    destination
       *           The procedure where the internal entry was resolved.
       * @param    iename
       *           The internal entry name. In case of triggers or external programs, use the 
       *           {@link #EXTERNAL_PROGRAM} and {@link #TRIGGER_PROCEDURE} constants.
       * @param    pname
       *           The external program name; will be non-null only when we are invoking an external
       *           program.
       * @param    function
       *           Flag indicating we are calling a user-defined function.
       * @param    persistentProc
       *           Flag indicating the external program was ran persistent.
       */
      public CalleeInfoImpl(boolean superCall,
                            Object  target,
                            Object  destination,
                            String  iename,
                            String  pname,
                            boolean function,
                            boolean persistentProc)
      {
         this.superCall      = superCall;
         this.target         = target;
         this.destination    = destination;
         this.iename         = iename;
         this.pname          = pname;
         this.function       = function;
         this.persistentProc = persistentProc;
      }
      
      /**
       * Called by {@link WorkArea#scopeStart()} - depending on the callee's configuration, it
       * populates the procedure system handles and maintains persistent procedures.
       */
      public void push()
      {
         WorkArea wa = work.get();
         
         if (pname != null && !wa.pMap.containsKey(target))
         {
            // this is an external procedure invocation...
            handle htarget = new handle(new ExternalProgramWrapper(target));
            ProcedureManager.addProcedure(htarget, pname, persistentProc);
         }

         if (!superCall)
         {
            if (!wa.thisProcedures.isEmpty())
            {
               wa.sourceProcedures.push(wa.thisProcedures.peek());
            }
            wa.targetProcedures.push(target);
         }

         pushInternalEntry(destination, iename, function);
      }
      
      /**
       * Called by {@link WorkArea#scopeFinished()} - depending on the callee's configuration, it
       * backs out the data pushed by {@link #push()}, so the state will be the same as before
       * the callee as invoked.
       */
      public void pop()
      {
         WorkArea wa = work.get();

         wa.internalEntryStack.pop();
         
         if (!superCall)
         {
            if (!wa.thisProcedures.isEmpty())
            {
               wa.sourceProcedures.pop();
            }

            wa.targetProcedures.pop();
         }
      }

      /** 
       * The internal entry name. In case of triggers or external programs, returns the 
       * {@link ProcedureManager#TRIGGER_PROCEDURE} and {@link ProcedureManager#EXTERNAL_PROGRAM} 
       * constants respectively.
       * 
       * @return   The internal entry name.
       */
      @Override
      public String getInternalEntryName()
      {
         return iename;
      }

      /**
       * The procedure where the internal entry was resolved.
       * 
       * @return   A valid instance.
       */
      @Override
      public Object getDestination()
      {
         return destination;
      }
      
      /** 
       * The external program name; will be non-null only when we are invoking an external 
       * program.
       * 
       * @return   A {@link String} representing the program name.
       */
      @Override 
      public String getExternalProgramName()
      {
         return pname;
      }
      
      /** 
       * Flag indicating a user-defined function is called. 
       *
       * @return   A boolean flag.
       */
      @Override
      public boolean isFunction()
      {
         return function;
      }
   }
   
   /**
    * Container for various procedure data.
    */
   private static class ProcedureData
   {
      /** The subscriptions to which this procedure is registered. */
      private final List<Subscription> subscriptions = new ArrayList<Subscription>();
      
      /** The list of super-procedures for this procedure, by external program instance. s*/
      private final Map<Object, ProcedureData> superProcedures = 
         new LinkedHashMap<Object, ProcedureData>();
      
      /** 
       * The map of name of functions defined IN handle in this external program, to their 
       * {@link Resolvable} instance. 
       */
      private final Map<String, Resolvable> inHandleFunctions = new HashMap<String, Resolvable>();
      
      /**
       * A handle instance for this procedure, so that other handles can be instantiated from it.
       */
      private handle phandle = null;
      
      /** The current search-mode for this external program. */
      private SearchMode searchMode = null;
      
      /** The private data of this procedure handle. */
      private character privateData = null;
      
      /** Implementation of ADM-DATA attribute. */
      private final character admData = new character();

      /** Implementation of UNIQUE-ID attribute. */
      private final integer uniqueID = new integer();
      
      /** Handle referencing the instantiating procedure. */
      private final handle instProc = new handle();
      
      /**
       * Set of {@link Scopeable} instances which need notification when the persistent procedure 
       * gets deleted.
       */
      private Set<Scopeable> scopeables = new HashSet<>();

      /**
       * Set of {@link Finalizable} instances which need notification when the persistent procedure 
       * gets deleted.
       */
      private Set<Finalizable> finalizables = new HashSet<>();

      /**
       * A stack holding the resources created in each top-level block.
       */
      private Deque<Set<WrappedResource>> staticResources = new ArrayDeque<>();
      
      /**
       * A value of CURRENT-WINDOW attribute of THIS-PROCEDURE handle
       */
      private handle currentWindow = new handle();
      
      /**
       * Create a new {@link ProcedureData} instance using the given 
       * procedure handle and search-mode.
       * 
       * @param    phandle
       *           The procedure handle.
       * @param    searchMode
       *           The search-mode. May be null.
       */
      ProcedureData(handle phandle, SearchMode searchMode)
      {
         this.phandle = new handle(phandle);
         this.searchMode = searchMode;
         this.uniqueID.assign(new integer(UniqueIdGenerator.getUniqueId(IdKind.PROCEDURE)));
      }
      
      /**
       * Check if the search-mode for this external program is SEARCH-SELF.
       * 
       * @return   See above.
       */
      boolean isSearchSelf()
      {
         return searchMode == null || searchMode == SearchMode.SEARCH_SELF;
      }
   }
   
   /**
    * Simple container that stores and returns a context-local instance of
    * the global work area.
    */
   private static class ContextContainer
   extends ContextLocal<WorkArea>
   {
      /**
       * Get the weight of this context-local var.
       * 
       * @return   Always {@link WeightFactor#LEVEL_5}.
       */
      @Override
      public WeightFactor getWeight()
      {
         return WeightFactor.LEVEL_5;
      };

      /**
       * Initializes the work area, the first time it is requested within a
       * new context. Registers the {@link WorkArea} instance with the
       * {@link TransactionManager}, for scope notifications.
       *
       * @return   The newly instantiated work area.
       */
      protected synchronized WorkArea initialValue()
      {
         WorkArea wa = new WorkArea();

         return wa;
      }
   }
}